@@ -27,19 +27,27 @@ class webull:
             'Accept-Encoding': 'gzip, deflate',
             'Content-Type': 'application/json',
             'platform': 'web',
-            'ver': '3.21.14',
+            'ver': '3.22.20',
             'User-Agent': '*',
+            'lzone': 'dc_core_r001',
             'did': self._get_did(),
         }
-        self._access_token = ''
+
+        #endpoints
+        self._urls = endpoints.urls()
+
+        #sessions
         self._account_id = ''
+        self._trade_token = ''
+        self._access_token = ''
         self._refresh_token = ''
         self._token_expire = ''
-        self._trade_token = ''
         self._uuid = ''
-        self._region_code = 6
+
+        #miscellaenous
         self._did = self._get_did()
-        self._urls = endpoints.urls()
+        self._region_code = 6
+        self.zone_var = 'dc_core_r001'
 
     def _get_did(self, path=''):
         '''
@@ -63,18 +71,19 @@ class webull:
             pickle.dump(did, open(filename, 'wb'))
         return did
 
-
-    def build_req_headers(self, include_trade_token=False, include_time=False):
+    def build_req_headers(self, include_trade_token=False, include_time=False, include_zone_var=True):
         '''
         Build default set of header params
         '''
         headers = self._headers
         headers['did'] = self._did
         headers['access_token'] = self._access_token
-        if include_trade_token:
+        if include_trade_token :
             headers['t_token'] = self._trade_token
-        if include_time:
+        if include_time :
             headers['t_time'] = str(round(time.time() * 1000))
+        if include_zone_var :
+            headers['lzone'] = self.zone_var
         return headers
 
 
@@ -94,19 +103,14 @@ class webull:
         password = ('wl_app-a&b@!423^' + password).encode('utf-8')
         md5_hash = hashlib.md5(password)
 
-        try:
-          validate_email(username)
-          accountType = 2 # email
-        except EmailNotValidError as _e:
-          accountType = 1 # phone
+        account_type = self.get_account_type(username)
 
         if device_name == '' :
             device_name = 'default_string'
 
         data = {
-            'accessQuestions': '[{"questionId":"' + str(question_id) + '", "answer":"' + str(question_answer) + '"}]',
             'account': username,
-            'accountType': accountType,
+            'accountType': str(account_type),
             'deviceId': self._did,
             'deviceName': device_name,
             'grade': 1,
@@ -116,12 +120,16 @@ class webull:
 
         if mfa != '' :
             data['extInfo'] = {
-                'codeAccountType': accountType,
+                'codeAccountType': account_type,
                 'verificationCode': mfa
             }
             headers = self.build_req_headers()
         else :
             headers = self._headers
+
+        if question_id != '' and question_answer != '' :
+            data['accessQuestions'] = '[{"questionId":"' + str(question_id) + '", "answer":"' + str(question_answer) + '"}]'
+
         response = requests.post(self._urls.login(), json=data, headers=headers)
         result = response.json()
         if 'accessToken' in result :
@@ -133,33 +141,25 @@ class webull:
         return result
 
     def get_mfa(self, username='') :
-        try:
-          validate_email(username)
-          accountType = 2 # email
-        except EmailNotValidError as _e:
-          accountType = 1 # phone
+        account_type = self.get_account_type()
 
         data = {'account': str(username),
-                'accountType': str(accountType),
+                'accountType': str(account_type),
                 'codeType': int(5)}
 
         response = requests.post(self._urls.get_mfa(), json=data, headers=self._headers)
         # data = response.json()
-        
+
         if response.status_code == 200 :
             return True
         else :
             return False
 
     def check_mfa(self, username='', mfa='') :
-        try:
-          validate_email(username)
-          accountType = 2 # email
-        except EmailNotValidError as _e:
-          accountType = 1 # phone
+        account_type = self.get_account_type(username)
 
         data = {'account': str(username),
-                'accountType': str(accountType),
+                'accountType': str(account_type),
                 'code': str(mfa),
                 'codeType': int(5)}
 
@@ -169,28 +169,39 @@ class webull:
         return data
 
     def get_security(self, username='') :
-        try:
-          validate_email(username)
-          accountType = 2 # email
-        except EmailNotValidError as _e:
-          accountType = 1 # phone
-
+        account_type = self.get_account_type(username)
         username = urllib.parse.quote(username)
 
-        response = requests.get(self._urls.get_security(username, accountType, self._region_code, 'PRODUCT_LOGIN'), headers=self._headers)
+        # seems like webull has a bug/stability issue here:
+        time = datetime.now().timestamp() * 1000
+        response = requests.get(self._urls.get_security(username, account_type, self._region_code, 'PRODUCT_LOGIN', time, 0), headers=self._headers)
         data = response.json()
+        if len(data) == 0 :
+            response = requests.get(self._urls.get_security(username, account_type, self._region_code, 'PRODUCT_LOGIN', time, 1), headers=self._headers)
+            data = response.json()
+
+        return data
+
+    def next_security(self, username='') :
+        account_type = self.get_account_type(username)
+        username = urllib.parse.quote(username)
+
+        # seems like webull has a bug/stability issue here:
+        time = datetime.now().timestamp() * 1000
+        response = requests.get(self._urls.next_security(username, account_type, self._region_code, 'PRODUCT_LOGIN', time, 0), headers=self._headers)
+        data = response.json()
+        if len(data) == 0 :
+            response = requests.get(self._urls.next_security(username, account_type, self._region_code, 'PRODUCT_LOGIN', time, 1), headers=self._headers)
+            print(response)
+            data = response.json()
 
         return data
 
     def check_security(self, username='', question_id='', question_answer='') :
-        try:
-          validate_email(username)
-          accountType = 2 # email
-        except EmailNotValidError as _e:
-          accountType = 1 # phone
+        account_type = self.get_account_type(username)
 
         data = {'account': str(username),
-                'accountType': str(accountType),
+                'accountType': str(account_type),
                 'answerList': [{'questionId': str(question_id), 'answer': str(question_answer)}],
                 'event': 'PRODUCT_LOGIN'}
 
@@ -259,7 +270,8 @@ class webull:
 
         response = requests.get(self._urls.account_id(), headers=headers)
         result = response.json()
-        if result['success']:
+        if result['success'] :
+            self.zone_var = str(result['data'][0]['rzone'])
             id = str(result['data'][0]['secAccountId'])
             self._account_id = id
             return id
@@ -309,7 +321,7 @@ class webull:
         data = self.get_account()
         return data['openOrders']
 
-    def get_history_orders(self, status='Cancelled', count=20):
+    def get_history_orders(self, status='All', count=20):
         '''
         Historical orders, can be cancelled or filled
         status = Cancelled / Filled / Working / Partially Filled / Pending / Failed / All
@@ -332,8 +344,8 @@ class webull:
 
         response = requests.post(self._urls.trade_token(), json=data, headers=headers)
         result = response.json()
-        if result['success']:
-            self._trade_token = result['data']['tradeToken']
+        if 'tradeToken' in result :
+            self._trade_token = result['tradeToken']
             return True
         else:
             return False
@@ -390,7 +402,7 @@ class webull:
         Place an order
 
         price: float (LMT / STP LMT Only)
-        action: BUY / SELL
+        action: BUY / SELL / SHORT
         ordertype : LMT / MKT / STP / STP LMT / STP TRAIL
         timeinforce:  GTC / DAY / IOC
         outsideRegularTradingHour: True / False
@@ -408,6 +420,7 @@ class webull:
         headers = self.build_req_headers(include_trade_token=True, include_time=True)
         data = {
             'action': action,
+            'comboType': 'NORMAL',
             'orderType': orderType,
             'outsideRegularTradingHour': outsideRegularTradingHour,
             'quantity': int(quant),
@@ -1000,7 +1013,7 @@ class webull:
         params = {'currentNewsId': Id, 'pageSize': items}
         return requests.get(self._urls.news(self.get_ticker(stock)), params=params, headers=headers).json()
 
-    def get_bars(self, stock=None, tId = None, interval='m1', count=1, extendTrading=0, timeStamp=None):
+    def get_bars(self, stock=None, tId=None, interval='m1', count=1, extendTrading=0, timeStamp=None):
         '''
         get bars returns a pandas dataframe
         params:
@@ -1038,6 +1051,42 @@ class webull:
             df.loc[to_datetime(datetime.fromtimestamp(int(row[0])).astimezone(time_zone))] = data
         return df.iloc[::-1]
 
+    def get_options_bars(self, derivativeId=None, interval='1m', count=1, direction=1, timeStamp=None):
+        '''
+        get bars returns a pandas dataframe
+        params:
+            derivativeId: to be obtained from option chain, eg option_chain[0]['call']['tickerId']
+            interval: 1m, 5m, 30m, 60m, 1d
+            count: number of bars to return
+            direction: 1 ignores {count} parameter & returns all bars on and after timestamp
+                       setting any other value will ignore timestamp & return latest {count} bars
+            timeStamp: If epoc timestamp is provided, return bar count up to timestamp. If not set default to current time.
+        '''
+        headers = self.build_req_headers()
+        if derivativeId is None:
+            raise ValueError('Must provide a derivative ID')
+
+        params = {'type': interval, 'count': count, 'direction': direction, 'timestamp': timeStamp}
+        df = DataFrame(columns=['open', 'high', 'low', 'close', 'volume', 'vwap'])
+        df.index.name = 'timestamp'
+        response = requests.get(self._urls.options_bars(derivativeId), params=params, headers=headers)
+        result = response.json()
+        time_zone = timezone(result[0]['timeZone'])
+        for row in result[0]['data'] :
+            row = row.split(',')
+            row = ['0' if value == 'null' else value for value in row]
+            data = {
+                'open': float(row[1]),
+                'high': float(row[3]),
+                'low': float(row[4]),
+                'close': float(row[2]),
+                'volume': float(row[6]),
+                'vwap': float(row[7])
+            }
+            #convert to a panda datetime64 which has extra features like floor and resample
+            df.loc[to_datetime(datetime.fromtimestamp(int(row[0])).astimezone(time_zone))] = data
+        return df.iloc[::-1]
+
     def get_calendar(self,stock=None, tId=None):
         '''
         There doesn't seem to be a way to get the times the market is open outside of the charts.
@@ -1114,7 +1163,7 @@ class webull:
                     rank = data['data']
         return rank
 
-    def get_watchlists(self):
+    def get_watchlists(self) :
         """
         get user watchlists
         :return:
@@ -1124,7 +1173,14 @@ class webull:
         response = requests.get(self._urls.portfolio_lists(), params=params, headers=headers)
         return response.json()['portfolioList']
 
+    def get_account_type(self, username='') :
+        try:
+            validate_email(username)
+            account_type = 2 # email
+        except EmailNotValidError as _e:
+            account_type = 1 # phone
 
+        return account_type
 
 ''' Paper support '''
 class paper_webull(webull):
@@ -1143,7 +1199,7 @@ class paper_webull(webull):
         headers = self.build_req_headers()
         response = requests.get(self._urls.paper_account_id(), headers=headers)
         result = response.json()
-        if result is not None and result[0] is not None:
+        if result is not None and len(result) > 0 and 'id' in result[0]:
             id = result[0]['id']
             self._account_id = id
             return id
@@ -1242,7 +1298,6 @@ class paper_webull(webull):
         result = response.json()
         return result
 
-
 if __name__ == '__main__':
     parser = argparse.ArgumentParser(description='Interface with Webull. Paper trading is not the default.')
     parser.add_argument('-p', '--use-paper', help='Use paper account instead.', action='store_true')
